import datetime
import sys
import unittest


class BaseTestCase(unittest.TestCase):
    def tap(self, out):
        sys.stderr.write("--- tap output start ---\n")
        for line in out.splitlines():
            sys.stderr.write(line + "\n")
        sys.stderr.write("---  tap output end  ---\n")


class TestCase(BaseTestCase):
    def assertOpenInterval(self, interval,
                           expectedId=None,
                           expectedStart=None,
                           expectedTags=None,
                           expectedAnnotation=None,
                           description="interval"):
        self.assertKeyExists(interval, "start", description, "{} does not contain a start date")
        self.assertKeyNotExists(interval, "end", description, "{} does contain an end date")

        return self.assertInterval(interval,
                                   expectedId=expectedId,
                                   expectedStart=expectedStart,
                                   expectedEnd=None,
                                   expectedTags=expectedTags,
                                   expectedAnnotation=expectedAnnotation,
                                   description=description)

    def assertClosedInterval(self, interval,
                             expectedId=None,
                             expectedStart=None,
                             expectedEnd=None,
                             expectedTags=None,
                             expectedAnnotation=None,
                             description="interval"):
        self.assertKeyExists(interval, "start", description, "{} does not contain a start date")
        self.assertKeyExists(interval, "end", description, "{} does not contain an end date")

        return self.assertInterval(interval,
                                   expectedId=expectedId,
                                   expectedStart=expectedStart,
                                   expectedEnd=expectedEnd,
                                   expectedTags=expectedTags,
                                   expectedAnnotation=expectedAnnotation,
                                   description=description)

    def assertInterval(self, interval,
                       expectedId=None,
                       expectedStart=None,
                       expectedEnd=None,
                       expectedTags=None,
                       expectedAnnotation=None,
                       description="interval"):
        if expectedId is not None:
            self.assertKeyExists(interval, "id", description, "{} does not contain an id")
            self.assertIntervalValue(interval,
                                     "id",
                                     expectedId,
                                     description,
                                     "{} of {} do not match (expected: '{}', actual: '{}')")

        if expectedStart:
            self.assertIntervalTimestamp(interval, "start", expectedStart, description)

        if expectedEnd:
            self.assertIntervalTimestamp(interval, "end", expectedEnd, description)

        if expectedTags:
            self.assertKeyExists(interval, "tags", description, "{} does not contain tags")
            self.assertIntervalValue(interval,
                                     "tags",
                                     expectedTags,
                                     description,
                                     "{} of {} do not match (expected: '{}', actual: '{}')")

        if expectedAnnotation:
            self.assertKeyExists(interval, "annotation", description, "{} is not annotated")
            self.assertIntervalValue(interval,
                                     "annotation",
                                     expectedAnnotation,
                                     description,
                                     "{} of {} do not match (expected: '{}', actual: '{}')")

    def assertKeyExists(self, interval, key, description, message):
        self.assertTrue(key in interval, message.format(description))

    def assertKeyNotExists(self, interval, key, description, message):
        self.assertFalse(key in interval, message.format(description))

    def assertIntervalTimestamp(self, interval, key, expected, description):
        if isinstance(expected, datetime.datetime):
            expected = "{:%Y%m%dT%H%M%SZ}".format(expected)

        self.assertIntervalValue(interval,
                                 key,
                                 expected,
                                 description,
                                 "{} time of {} does not match (expected: '{}', actual: '{}')")

    def assertIntervalValue(self, interval, key, expected, description, message):
        actual = interval[key]

        if isinstance(actual, list):
            actual.sort()
            expected.sort()
            self.assertSequenceEqual(actual,
                                     expected,
                                     message.format(key, description, expected, actual))
        else:
            self.assertEqual(actual,
                             expected,
                             message.format(key, description, expected, actual))



from __future__ import division

import atexit
import functools
import os
import signal
import sys
import tempfile
from queue import Queue, Empty
from subprocess import Popen, PIPE, STDOUT
from threading import Thread
from time import sleep

try:
    import simplejson as json
except ImportError:
    import json
from .exceptions import CommandError, TimeoutWaitingFor

ON_POSIX = "posix" in sys.builtin_module_names

# Directory relative to basetest module location
CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

# Location of binary files (usually the src/ folder)
BIN_PREFIX = os.path.abspath(
    os.path.join(CURRENT_DIR, "..", "..", "src")
)

# Default location of test certificates
DEFAULT_CERT_PATH = os.path.abspath(
    os.path.join(CURRENT_DIR, "..", "test_certs")
)

# Default location of test extensions
DEFAULT_EXTENSION_PATH = os.path.abspath(
    os.path.join(CURRENT_DIR, "..", "test_extensions")
)


# Environment flags to control skipping of timew tests
TIMEW_SKIP = os.environ.get("TIMEW_SKIP", False)
# Environment flags to control use of PATH or in-tree binaries
TIMEW_USE_PATH = os.environ.get("TIMEW_USE_PATH", False)

UUID_REGEXP = ("[0-9A-Fa-f]{8}-" + ("[0-9A-Fa-f]{4}-" * 3) + "[0-9A-Fa-f]{12}")


def timew_binary_location(cmd="timew"):
    """ ../src/ is used by default."""
    return os.path.join(BIN_PREFIX, cmd)


def binary_location(cmd, USE_PATH=False):
    """ ../src/ is used by default."""
    return os.path.join(BIN_PREFIX, cmd)


def wait_condition(cond, timeout=1, sleeptime=.01):
    """Wait for condition to return anything other than None"""
    # NOTE Increasing sleeptime can dramatically increase testsuite runtime
    # It also reduces CPU load significantly
    if timeout is None:
        timeout = 1

    if timeout < sleeptime:
        print("Warning, timeout cannot be smaller than", sleeptime)
        timeout = sleeptime

    # Max number of attempts until giving up
    tries = int(timeout / sleeptime)

    for i in range(tries):
        val = cond()

        if val is not None:
            break

        sleep(sleeptime)

    return val


def wait_process(pid, timeout=None):
    """Wait for process to finish"""
    def process():
        try:
            os.kill(pid, 0)
        except OSError:
            # Process is dead
            return True
        else:
            # Process is still ticking
            return None

    return wait_condition(process, timeout)


def _queue_output(arguments, pidq, outputq):
    """Read/Write output/input of given process.
    This function is meant to be executed in a thread as it may block
    """
    kwargs = arguments["process"]
    input = arguments["input"]

    try:
        proc = Popen(**kwargs)
    except OSError as e:
        # pid None is read by the main thread as a crash of the process
        pidq.put(None)

        outputq.put((
            "",
            ("Unexpected exception caught during execution: '{0}' . ".format(e)),
            255))  # false exitcode

        return

    # Put the PID in the queue for main process to know.
    pidq.put(proc.pid)

    # Send input and wait for finish
    out, err = proc.communicate(input)

    if sys.version_info > (3,):
        out, err = out.decode("utf-8"), err.decode("utf-8")

    # Give the output back to the caller
    outputq.put((out, err, proc.returncode))


def _retrieve_output(thread, timeout, queue, thread_error):
    """Fetch output from binary subprocess queues"""
    # Try to join the thread on failure abort
    thread.join(timeout)
    if thread.is_alive():
        # Join should have killed the thread. This is unexpected
        raise TimeoutWaitingFor(thread_error + ". Unexpected error")

    # Thread died so we should have output
    try:
        # data = (stdout, stderr, exitcode)
        data = queue.get(timeout=timeout)
    except Empty:
        data = TimeoutWaitingFor("streams from program")

    return data


def _get_output(arguments, timeout=None):
    """Collect output from the subprocess without blocking the main process if
    subprocess hangs.
    """
    # NOTE Increase this value if tests fail with None being received as
    # stdout/stderr instead of the expected content
    output_timeout = 0.1  # seconds

    pidq = Queue()
    outputq = Queue()

    t = Thread(target=_queue_output, args=(arguments, pidq, outputq))
    t.daemon = True
    t.start()

    try:
        pid = pidq.get(timeout=timeout)
    except Empty:
        pid = None

    # Process crashed or timed out for some reason
    if pid is None:
        return _retrieve_output(t, output_timeout, outputq, "Program to start")

    # Wait for process to finish (normal execution)
    state = wait_process(pid, timeout)

    if state:
        # Process finished
        return _retrieve_output(t, output_timeout, outputq, "Program thread to join")

    # If we reach this point we assume the process got stuck or timed out
    for sig in (signal.SIGABRT, signal.SIGTERM, signal.SIGKILL):
        # Start with lower signals and escalate if process ignores them
        try:
            os.kill(pid, signal.SIGABRT)
        except OSError as e:
            # 3 means the process finished/died between last check and now
            if e.errno != 3:
                raise

        # Wait for process to finish (should die/exit after signal)
        state = wait_process(pid, timeout)

        if state:
            # Process finished
            return _retrieve_output(t, output_timeout, outputq, "Program to die")

    # This should never happen, but in case something goes really awry...
    raise OSError("Program stopped responding and couldn't be killed")


def run_cmd_wait(cmd, input=None, stdout=PIPE, stderr=PIPE,
                 merge_streams=False, env=os.environ, timeout=None):
    """Run a subprocess and wait for it to finish"""

    if input is None:
        stdin = None
    else:
        stdin = PIPE

    if merge_streams:
        stderr = STDOUT
    else:
        stderr = PIPE

    arguments = {
        "process": {
            "args": cmd,
            "stdin": stdin,
            "stdout": stdout,
            "stderr": stderr,
            "bufsize": -1,
            "close_fds": ON_POSIX,
            "env": env,
        },
        "input": input,
    }
    out, err, exit = _get_output(arguments, timeout)

    if merge_streams:
        if exit != 0:
            raise CommandError(cmd, exit, out)
        else:
            return exit, out
    else:
        if exit != 0:
            raise CommandError(cmd, exit, out, err)
        else:
            return exit, out, err


def run_cmd_wait_nofail(*args, **kwargs):
    """Same as run_cmd_wait but silence the exception if it happens"""
    try:
        return run_cmd_wait(*args, **kwargs)
    except CommandError as e:
        return e.code, e.out, e.err


def memoize(obj):
    """Keep an in-memory cache of function results given its inputs"""
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer


try:
    from shutil import which
    which = memoize(which)
except ImportError:
    # NOTE: This is shutil.which backported from python-3.3.3
    @memoize
    def which(cmd, mode=os.F_OK | os.X_OK, path=None):
        """Given a command, mode, and a PATH string, return the path which
        conforms to the given mode on the PATH, or None if there is no such
        file.

        `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
        of os.environ.get("PATH"), or can be overridden with a custom search
        path.

        """
        # Check that a given file can be accessed with the correct mode.
        # Additionally check that `file` is not a directory, as on Windows
        # directories pass the os.access check.
        def _access_check(fn, mode):
            return (os.path.exists(fn) and os.access(fn, mode) and
                    not os.path.isdir(fn))

        # If we're given a path with a directory part, look it up directly
        # rather than referring to PATH directories. This includes checking
        # relative to the current directory, e.g. ./script
        if os.path.dirname(cmd):
            if _access_check(cmd, mode):
                return cmd
            return None

        if path is None:
            path = os.environ.get("PATH", os.defpath)
        if not path:
            return None
        path = path.split(os.pathsep)

        if sys.platform == "win32":
            # The current directory takes precedence on Windows.
            if os.curdir not in path:
                path.insert(0, os.curdir)

            # PATHEXT is necessary to check on Windows.
            pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
            # See if the given file matches any of the expected path
            # extensions. This will allow us to short circuit when given
            # "python.exe". If it does match, only test that one, otherwise we
            # have to try others.
            if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
                files = [cmd]
            else:
                files = [cmd + ext for ext in pathext]
        else:
            # On other platforms you don't have things like PATHEXT to tell you
            # what file suffixes are executable, so just pass on cmd as-is.
            files = [cmd]

        seen = set()
        for dir in path:
            normdir = os.path.normcase(dir)
            if normdir not in seen:
                seen.add(normdir)
                for thefile in files:
                    name = os.path.join(dir, thefile)
                    if _access_check(name, mode):
                        return name
        return None


def parse_datafile(file):
    """Parse .data files, treating files as JSON"""
    data = []
    with open(file) as fh:
        for line in fh:
            line = line.rstrip("\n")

            # Turn [] strings into {} to be treated properly as JSON hashes
            if line.startswith("[") and line.endswith("]"):
                line = "{" + line[1:-1] + "}"

            if line.startswith("{"):
                data.append(json.loads(line))
            else:
                data.append(line)
    return data


def mkstemp(data):
    """Create a temporary file that is removed at process exit"""
    def rmtemp(name):
        try:
            os.remove(name)
        except OSError:
            pass

    f = tempfile.NamedTemporaryFile(delete=False)
    f.write(data)
    f.close()

    # Ensure removal at end of python session
    atexit.register(rmtemp, f.name)

    return f.name


def mkstemp_exec(data):
    """Create a temporary executable file that is removed at process exit"""
    name = mkstemp(data)
    os.chmod(name, 0o755)

    return name



#!/usr/bin/env python3

import os

import argparse
import matplotlib.pyplot as plt
import numpy as np

parser = argparse.ArgumentParser(description='Display performance plots.')
parser.add_argument('output_dir', help='directory containing the measurement files')
parser.add_argument('commands', nargs='*', type=str)

args = parser.parse_args()

output_directory = args.output_dir
commands = args.commands

plt.axes([0.1, 0.1, 0.6, 0.75])
plt.xlabel("# database entries")
plt.ylabel("time [s]")
plt.title(output_directory)

for filename in os.listdir(output_directory):
    if filename.endswith(".log"):
        cmd = "-".join(filename.split('-')[1:-1])
    else:
        continue

    if len(commands) == 0 or cmd in commands:
        try:
            x, y = np.loadtxt(os.path.join(output_directory, filename),
                              delimiter='\t',
                              usecols=(1, 2),
                              unpack=True)
            plt.plot(x, y, label=cmd, marker=".", linestyle='-')
        except ValueError as e:
            print("Invalid file: {} {}".format(filename, e))

plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0., ncol=2)
plt.show()



#!/usr/bin/env python3

import json
import sys

# Skip the configuration settings.
for line in sys.stdin:
    if line == '\n':
        break

# Extract the JSON.
doc = ''
for line in sys.stdin:
    doc += line

total_active_time = 0

j = json.loads(doc)
for object in j:
    line = '"{}","{}"'.format(object['start'], (object['end'] if 'end' in object else ''))

    if 'tags' in object:
        for tag in object['tags']:
            line += ',"{}"'.format(tag)

    print(line)



import atexit
import datetime
import json
import os
import shlex
import shutil
import tempfile
import unittest

from .exceptions import CommandError
from .utils import run_cmd_wait, run_cmd_wait_nofail, which, timew_binary_location, DEFAULT_EXTENSION_PATH


class Timew(object):
    """
    Manage a Timewarrior instance

    A temporary folder is used as data store of timewarrior.

    A timew client should not be used after being destroyed.
    """
    DEFAULT_TIMEW = timew_binary_location()

    def __init__(self, timew=DEFAULT_TIMEW):
        """
        Initialize a timewarrior (client).
        The program runs in a temporary folder.

        :arg timew: Timewarrior binary to use as client (defaults: timew in PATH)
        """
        self.timew = timew

        # Used to specify what command to launch (and to inject faketime)
        self._command = [self.timew]

        # Configuration of the isolated environment
        self._original_pwd = os.getcwd()
        self.datadir = tempfile.mkdtemp(prefix="timew_")
        self.timewrc = os.path.join(self.datadir, "timewarrior.cfg")
        self.extdir = os.path.join(self.datadir, "extensions")

        # Ensure any instance is properly destroyed at session end
        atexit.register(lambda: self.destroy())

        self.reset_env()

    def add_default_extension(self, filename):
        """Add default extension to current instance"""
        if not os.path.isdir(self.extdir):
            os.mkdir(self.extdir)

        extfile = os.path.join(self.extdir, filename)
        if os.path.isfile(extfile):
            raise "{} already exists".format(extfile)

        shutil.copy(os.path.join(DEFAULT_EXTENSION_PATH, filename), extfile)

    def __repr__(self):
        txt = super(Timew, self).__repr__()
        return "{0} running from {1}>".format(txt[:-1], self.datadir)

    def __call__(self, *args, **kwargs):
        """aka t = Timew() ; t() which is now an alias to t.runSuccess()"""
        return self.runSuccess(*args, **kwargs)

    def reset_env(self):
        """Set a new environment derived from the one used to launch the test"""
        # Copy all env variables to avoid clashing subprocess environments
        self.env = os.environ.copy()

        # As well as TIMEWARRIORDB
        self.env["TIMEWARRIORDB"] = self.datadir

        # As well as MANPATH, so that the help tests can find the
        # uninstalled man pages
        parts = self.timew.split(os.path.sep)[0:-2]
        parts.append("doc")
        self.env["MANPATH"] = os.path.sep.join(parts)

    def config(self, var, value):
        """Run setup `var` as `value` in timew config"""
        cmd = (self.timew, ":yes", "config", var, value)
        return run_cmd_wait(cmd, env=self.env)

    @staticmethod
    def _create_exclusion_interval(interval):
        if not isinstance(interval, tuple):
            raise TypeError("Please specify interval(s) as a tuple or a list of tuples")

        if interval[0] is not None and not isinstance(interval[0], datetime.time):
            raise TypeError("Start date must be a datetime.time but is {}".format(type(interval[0])))

        if interval[1] is not None and not isinstance(interval[1], datetime.time):
            raise TypeError("End date must be a datetime.time but is {}".format(type(interval[1])))

        if interval[0] is None:
            return "<{:%H:%M:%S}".format(interval[1])

        if interval[1] is None:
            return ">{:%H:%M:%S}".format(interval[0])

        if interval[0] > interval[1]:
            return "<{:%H:%M:%S} >{:%H:%M:%S}".format(interval[1], interval[0])

        return "{:%H:%M:%S}-{:%H:%M:%S}".format(interval[0], interval[1])

    def configure_exclusions(self, intervals):
        if isinstance(intervals, list):
            exclusion = " ".join([self._create_exclusion_interval(interval) for interval in intervals])

        else:
            exclusion = self._create_exclusion_interval(intervals)

        self.config("exclusions.monday", exclusion)
        self.config("exclusions.tuesday", exclusion)
        self.config("exclusions.wednesday", exclusion)
        self.config("exclusions.thursday", exclusion)
        self.config("exclusions.friday", exclusion)
        self.config("exclusions.saturday", exclusion)
        self.config("exclusions.sunday", exclusion)

    def del_config(self, var):
        """Remove `var` from timew config"""
        cmd = (self.timew, ":yes", "config", var)
        return run_cmd_wait(cmd, env=self.env)

    @property
    def timewrc_content(self):
        """Returns the contents of the timewrc file."""

        with open(self.timewrc, "r") as f:
            return f.readlines()

    def export(self, export_filter=None):
        """Run "timew export", return JSON array of exported intervals."""
        if export_filter is None:
            export_filter = ""

        code, out, err = self.runSuccess("{0} export".format(export_filter))

        return json.loads(out)

    @staticmethod
    def _split_string_args_if_string(args):
        """
        Helper function to parse and split into arguments a single string argument.
        The string is literally the same as if written in the shell.
        """
        # Enable nicer-looking calls by allowing plain strings
        if isinstance(args, str):
            args = shlex.split(args)

        return args

    def runSuccess(self, args="", input=None, merge_streams=False, timeout=5):
        """
        Invoke timew with given arguments and fail if exit code != 0

        Use runError if you want exit_code to be tested automatically and
        *not* fail if program finishes abnormally.

        If you wish to pass instructions to timew such as confirmations or other
        input via stdin, you can do so by providing an input string.
        Such as input="y\ny\n".

        If merge_streams=True stdout and stderr will be merged into stdout.

        timeout = number of seconds the test will wait for every timew call.
        Defaults to 1 second if not specified. Unit is seconds.

        Returns (exit_code, stdout, stderr) if merge_streams=False
                (exit_code, output) if merge_streams=True
        """
        # Create a copy of the command
        command = self._command[:]

        args = self._split_string_args_if_string(args)
        command.extend(args)

        output = run_cmd_wait_nofail(command, input,
                                     merge_streams=merge_streams,
                                     env=self.env,
                                     timeout=timeout)

        if output[0] != 0:
            raise CommandError(command, *output)

        return output

    def runError(self, args=(), input=None, merge_streams=False, timeout=5):
        """
        Invoke timew with given arguments and fail if exit code == 0

        Use runSuccess if you want exit_code to be tested automatically and
        *fail* if program finishes abnormally.

        If you wish to pass instructions to timew such as confirmations or other
        input via stdin, you can do so by providing an input string.
        Such as input="y\ny\n".

        If merge_streams=True stdout and stderr will be merged into stdout.

        timeout = number of seconds the test will wait for every timew call.
        Defaults to 1 second if not specified. Unit is seconds.

        Returns (exit_code, stdout, stderr) if merge_streams=False
                (exit_code, output) if merge_streams=True
        """
        # Create a copy of the command
        command = self._command[:]

        args = self._split_string_args_if_string(args)
        command.extend(args)

        output = run_cmd_wait_nofail(command, input,
                                     merge_streams=merge_streams,
                                     env=self.env,
                                     timeout=timeout)

        # output[0] is the exit code
        if output[0] == 0 or output[0] is None:
            raise CommandError(command, *output)

        return output

    def destroy(self):
        """Cleanup the data folder and release server port for other instances"""
        try:
            shutil.rmtree(self.datadir)
        except OSError as e:
            if e.errno == 2:
                # Directory no longer exists
                pass
            else:
                raise

        # Prevent future reuse of this instance
        self.runSuccess = self.__destroyed
        self.runError = self.__destroyed

        # self.destroy will get called when the python session closes.
        # If self.destroy was already called, turn the action into a noop
        self.destroy = lambda: None

    def __destroyed(self, *args, **kwargs):
        raise AttributeError("Program instance has been destroyed. "
                             "Create a new instance if you need a new client.")

    def faketime(self, faketime=None):
        """
        Set a faketime using libfaketime that will affect the following command calls.

        If faketime is None, faketime settings will be disabled.
        """
        cmd = which("faketime")
        if cmd is None:
            raise unittest.SkipTest("libfaketime/faketime is not installed")

        if self._command[0] == cmd:
            self._command = self._command[3:]

        if faketime is not None:
            # Use advanced time format
            self._command = [cmd, "-f", faketime] + self._command
